// expression parser.
// author: Dmitry Boyarintsev
unit ExprParse;

{$mode delphi}

interface

uses
  Classes, SysUtils, TextParsingUtils;

type
  { TExpression }

  TExpression = class(TObject)
    oper      : AnsiString;
    constructor Create(const aoper: AnsiString);
  end;

  { TIdentExp }

  TIdentExp = class(TExpression)
    ident     : AnsiString;
    constructor Create(const aident: AnsiString);
  end;

  { TExpression }

  TUnaryPrefixExp = class(TExpression)
    right   : TExpression;
    rightid : AnsiString;
  end;

  { TBinExp }

  TBinExp = class(TExpression)
    leftid  : AnsiString;
    rightid : AnsiString;
    left    : TExpression;
    right   : TExpression;
    destructor Destroy; override;
  end;

  TStatement = class
    Next  : TStatement;
  end;

  TExpressionStatement = class(TStatement)
    Expression  : TExpression;
  end;

  TAssignStatement = class(TStatement)
    Left    : TExpression;
    Right   : TExpression;
  end;

  TIfStatement = class(TStatement)
    Condition     : TExpression;
    ThenStatement : TStatement;
    ElseStatement : TStatement;
  end;

  TWhileStatement = class(TStatement)
    Condition : TExpression;
    Body      : TStatement;
  end;

  TExpTokenKind = (etk_None, etk_Identifier, etk_Symbol, etk_Number,
    etk_ExpStart, etk_ExpEnd);

  { TExpParser }

  TExpParser = class(TObject)
  protected
    procedure Error(const Cmt: String);
  public
    index       : Integer;
    ErrorState  : AnsiString;
    Text        : AnsiString;
    Token       : AnsiString;
    TokenPos    : Integer;
    Kind        : TExpTokenKind;
    ExpEnd      : TStringList;
    constructor Create;
    destructor Destroy; override;
    function NextToken: Boolean; virtual;
  end;

function ParseExpression(parser: TExpParser): TExpression; overload;
function ParseExpression(const expstr: AnsiString): TExpression; overload;

implementation

const
  TestSymChars = ['+','/','-','*',':','=','<','>'];

{
Logical negation: not
Multiplicative: * / div mod and
Additive: + - or
Relational: = <> < > <= >= in
}

{ TExpParser }

procedure TExpParser.Error(const Cmt:String);
begin
  ErrorState:=cmt;
end;

constructor TExpParser.Create;
begin
  ExpEnd:=TStringList.Create;
end;

destructor TExpParser.Destroy;
begin
  ExpEnd.Free;
  inherited Destroy;
end;

function isIdentSymbol(const s: String): Boolean;
begin
  Result:=(s='not') or
          (s='mod') or (s='div') or (s='and') or
          (s='as') or (s='is') or (s='or') or (s='xor') or
          (s='in');
end;

function TExpParser.NextToken:Boolean;
var
  i : integer;
begin
  i:=index;
  ScanWhile(Text, i, SpaceEolnChars);
  Result:=i<=length(Text);
  if not Result then begin
    Token:='';
    Kind:=etk_None;
    index:=i;
    Exit;
  end;

  Result:=True;
  TokenPos:=i;
  if Text[i] in NumericChars then begin
    Kind:=etk_Number;
    Token:=ScanWhile(Text,i, NumericChars);
  end else if Text[i] in AlphabetChars then begin
    Kind:=etk_Identifier;
    Token:=ScanWhile(Text,i, AlphaNumChars);
    if isIdentSymbol(Token) then Kind:=etk_Symbol;
    if ExpEnd.IndexOf(Token)>=0 then Kind:=etk_ExpEnd;
  end else if Text[i] in TestSymChars then begin
    Kind:=etk_Symbol;
    Token:=ScanWhile(Text,i, TestSymChars);
    if Token=':=' then Kind:=etk_ExpEnd;
  end else begin
    Kind:=etk_None;
    case Text[i] of
      '(': Kind:=etk_ExpStart;
      ')',';': Kind:=etk_ExpEnd;
    else
      Result:=False;
    end;
    if Result then begin
      Token:=Text[i];
      inc(i);
    end else
      Error('InvalidChar '+ Text[i]);
  end;
  index:=i;
end;

procedure PeekExp(v: TStringList; var st: String; var exp: TExpression);
begin
  if Assigned(v) and (v.Count>0) then begin
    st:=v[v.Count-1];
    exp:=TExpression(v.Objects[v.Count-1]);
  end else begin
    st:='';
    exp:=nil;
  end;
end;

procedure PopExp(v: TStringList; var st: String; var exp: TExpression);
begin
  PeekExp(v, st, exp);
  if v.Count>0 then v.Delete(v.Count-1)
end;

function PeekStr(v: TStringList): String;
var
  e : TExpression;
begin
  PeekExp(v, Result, e);
end;

function PopStr(v: TStringList): String;
var
  e : TExpression;
begin
  PeekExp(v, Result, e);
  if v.Count>0 then v.Delete(v.Count-1);
end;

function isOpPrefix(const s: String): Boolean;
begin
  Result:=s='not';
end;

function OpLevel(const s: String): Integer;
begin
  if (s='not') then Result:=4
  else if (s='*') or (s='/') or (s='div') or (s='mod') or (s='and') then Result:=3
  else if (s='+') or (s='-') or (s='or') or (s='xor')then Result:=2
  else if (s='=') or (s='<>') or (s ='<') or (s='>') or (s='<=') or (s='>=') or (s='in') then Result:=1
  else Result:=0;
end;

function CreateBinExpr(const leftid, rightid: AnsiString; leftexp, rightexp: TExpression; const op: AnsiString): TExpression;
var
  b: TBinExp;
begin
  b:=TBinExp.Create(op);
  if Assigned(leftexp) then b.left:=leftexp else b.leftid:=leftid;
  if Assigned(rightexp) then b.right:=rightexp else b.rightid:=rightid;
  Result:=b;
end;

function CreatePrefixUnary(const rightid: AnsiString; rightexp: TExpression; const op: AnsiString): TExpression;
var
  b: TUnaryPrefixExp;
begin
  b:=TUnaryPrefixExp.Create(op);
  if Assigned(rightexp) then b.right:=rightexp else b.rightid:=rightid;
  Result:=b;
end;


function ParseExpression(parser: TExpParser): TExpression;
var
  idstack : TStringList;
  opstack : TStringList;
  op      : String;
  tempop  : String;
  id1,id2 : String;
  e1, e2  : TExpression;
  i       : Integer;
  prefix  : TUnaryPrefixExp;

  procedure PopAndPushOperator;
  begin
    tempop:=PopStr(opstack);
    if not isOpPrefix(tempop) then begin
      PopExp(idstack, id2, e2);
      PopExp(idstack, id1, e1);
      idstack.AddObject('', CreateBinExpr(id1, id2, e1, e2, tempop));
    end else begin
      PopExp(idstack, id1, e1);
      idstack.AddObject('', CreatePrefixUnary(id1, e1, tempop));
    end;
  end;

begin
  idstack := TStringList.Create;
  opstack := TStringList.Create;

  try
    while parser.NextToken do begin
      case parser.Kind of
        etk_ExpStart:
        begin
          e1:=ParseExpression(parser);
          if Assigned(e1) then idstack.AddObject('', e1)
          else Break;
        end;
        etk_ExpEnd:
          Break;
        etk_Identifier, etk_Number:
          idstack.Add(parser.Token);
        etk_Symbol: begin
          op:=parser.Token;
          if opstack.Count=0 then
            opstack.Add(op)
          else begin
            tempop:=PeekStr(opstack);
            while OpLevel(tempop)>=OpLevel(op) do begin
              PopAndPushOperator;
              tempop:=PeekStr(opstack);
            end;
            opstack.Add(op);
          end;
        end;
      end;
    end;

    while opstack.Count>0 do
      PopAndPushOperator;

    if idstack.Count=1 then begin
      Result:=TExpression(idstack.Objects[0]);
      if Result=nil then
        Result:=TIdentExp.Create(idstack[0]);
    end else begin
      for i:=0 to idstack.Count-1 do idstack.Objects[i].Free;
      Result:=nil;
    end;

  finally
    idstack.Free;
    opstack.Free;
  end;

end;

function ParseExpression(const expstr: AnsiString): TExpression;
var
  p : TExpParser;
begin
  p := TExpParser.Create;
  p.Text:=expstr;
  p.index:=1;
  Result:=ParseExpression(p);
  p.Free;
end;

{ TExpression }

constructor TExpression.Create(const aoper:AnsiString);
begin
  oper:=aoper;
end;

{ TBinExp }

destructor TBinExp.Destroy;
begin
  if Assigned(left) then left.Free;
  if Assigned(right) then right.Free;
  inherited Destroy;
end;

{ TIdentExp }

constructor TIdentExp.Create(const aident:AnsiString);
begin
  inherited Create('');
  ident:=aident;
end;

end.

